
const
	http = require('http'),
	events = require('events')

const
	___error = require('io.error')

const
	___native = require(`./binaries/uws_${process.platform}_${process.versions.modules}`)

const
	WEB_SOCKET_OPEN = 1,
	WEB_SOCKET_CLOSED = 0,
	WEB_SOCKET_OPCODE_TEXT = 1,
	WEB_SOCKET_OPCODE_BINARY = 2,
	WEB_SOCKET_OPCODE_PING = 9,

	PER_MESSAGE_DEFLATE = 1,
	DEFAULT_PAYLOAD_LIMIT = 16777216

const
	___noop = () => {}

const
	___native_client_group = ___native
		.client
		.group
		.create(
			0, 
			DEFAULT_PAYLOAD_LIMIT
		)

const
	___onClientGroupConnection = (external) => {

		const
			socket = ___native.getUserData(external)

		socket.___external = external

		socket.___internalOnOpen()

	}

const
	___onClientGroupDisconnection = (external, code, message, socket) => {

		socket.___external = null

		process.nextTick(() => socket.___internalOnClose(code, message))

		___native.clearUserData(external)

	}

const
	___onClientGroupMessage = (message, socket) => {

		socket.___internalOnMessage(message)

	}

const
	___onClientGroupError = (socket) => {

		process.nextTick(() => socket.___internalOnError(___error('io.uws/UWS_CLIENT_CONNECTON_ERROR')))

	}

___native
	.setNoop(___noop)

___native
	.client
	.group
	.onConnection(
		___native_client_group,
		(external) => ___onClientGroupConnection(external)
	)

___native
	.client
	.group
	.onDisconnection(
		___native_client_group, 
		(external, code, message, socket) => ___onClientGroupDisconnection(external, code, message, socket)
	)

___native
	.client
	.group
	.onMessage(
		___native_client_group, 
		(message, socket) => ___onClientGroupMessage(message, socket)
	)

___native
	.client
	.group
	.onError(
		___native_client_group, 
		(socket) => ___onClientGroupError(socket)
	)

class WebSocket {

	constructor (external) {

		this.___external = external

		this.___internalOnMessage = ___noop
		this.___internalOnClose = ___noop

	}

	on (event, callback) {

		if (event === 'message') {

			this.___internalOnMessage = callback

		} else if (event === 'close') {

			this.___internalOnClose = callback

		} else if (event === 'open') {

			this.___internalOnOpen = callback

		} else if (event === 'error' && this instanceof Client) {

			this.___internalOnError = callback

		}

		return this

	}

	removeAllListeners (event) {

		if (event === 'message') {

			this.___internalOnMessage = ___noop

		}

		if (event === 'close') {

			this.___internalOnClose = ___noop

		}

		return this

	}

	removeListener (event, callback) {

		if (event === 'message' && this.___internalOnMessage === callback) {

			this.___internalOnMessage = ___noop

		} else if (event === 'close' && this.___internalOnClose === callback) {

			this.___internalOnClose = ___noop

		}

		return this

	}

	get readyState () {

		return this.___external ? WEB_SOCKET_OPEN : WEB_SOCKET_CLOSED

	}

	get address () {

		const
			address = this.___external ? ___native.getAddress(this.___external) : new Array(3)

		return {
			remotePort: address[0],
			remoteAddress: address[1],
			remoteFamily: address[2]
		}

	}

	terminate () {

		if (this.___external) {

			___native
				.server
				.terminate(this.___external)

			this.___external = null

		}

	}

	send (message, callback) {

		if (!this.___external) {

			return callback && callback(new Error('not opened'))

		}

		___native
			.server
			.send(
				this.___external, 
				message, 
				WEB_SOCKET_OPCODE_TEXT, 
				callback && process.nextTick(callback) || undefined
			)

	}

	close (code, data) {

		if (!this.___external) {

			return 

		}

		___native
			.server
			.close(
				this.___external, 
				code, 
				data
			)

		this.___external = null

	}

}

class Client extends WebSocket {

	constructor (uri) {

		super(null)

		this.___internalOnOpen = ___noop
		this.___internalOnError = ___noop

		___native.connect(
			___native_client_group, 
			uri, 
			this
		)

	}

	close (code, data) {

		if (!this.___external) {

			return

		}

		___native
			.client
			.close(
				this.___external, 
				code, 
				data
			)

		this.___external = null

	}

	terminate () {

		if (!this.___external) {

			return

		}

		___native
			.client
			.terminate(this.___external)

		this.___external = null

	}

	send (message, callback) {

		if (!this.___external) {

			return callback && callback(new Error('not opened'))

		}

		___native
			.client
			.send(
				this.___external, 
				message, 
				WEB_SOCKET_OPCODE_TEXT, 
				callback && process.nextTick(callback) || undefined
			)

	}

}

class Server extends events {

	constructor (options) {

		super()

		this.___http = new http.Server()

		this.___http

			.on('error', (error) => this.___onHttpError(error))
			.on('request', (req, socket) => this.___onHttpRequest(req, socket))
			.on('upgrade', (req, socket, head) => this.___onHttpUpgrade(req, socket, head))
			.on('listening', () => this.___onHttpListening())
			.listen(options.port, options.host || '127.0.0.1')

		this.___native_server_group = ___native
			.server
			.group
			.create(PER_MESSAGE_DEFLATE, DEFAULT_PAYLOAD_LIMIT)

		___native
			.server
			.group
			.onDisconnection(
				this.___native_server_group, 
				(external, code, message, socket) => this.___onNativeSocketDisconnection(external, code, message, socket)
			)

		___native
			.server
			.group
			.onMessage(
				this.___native_server_group, 
				(message, socket) => this.___onNativeSocketMessage(message, socket)
			)

		___native
			.server
			.group
			.onConnection(
				this.___native_server_group, 
				(external) => this.___onNativeSocketConnection(external)
			)

	}

	___onHttpError (error) {

		this.emit('error', error)

	}

	___onHttpRequest (req, socket) {

		socket.end()

	}

	___onHttpUpgrade (req, socket, head) {

		const
			sec_websocket_key = req.headers['sec-websocket-key'],
			sec_websocket_extensions = req.headers['sec-websocket-extensions'],
			sec_websocket_protocol = req.headers['sec-websocket-protocol']

		if (!socket._handle || !sec_websocket_key || sec_websocket_key.length !== 24) {

			return socket.destroy()

		}

		socket.setNoDelay(true)

		const
			native_transfer = ___native.transfer(socket._handle.fd === -1 ? socket._handle : socket._handle.fd, null)

		___native.upgrade(
			this.___native_server_group, 
			native_transfer, 
			sec_websocket_key, 
			sec_websocket_extensions, 
			sec_websocket_protocol
		)

		socket.destroy()

	}

	___onHttpListening () {

		this.emit('listening')

	}

	___onNativeSocketConnection (external) {

		const 
			socket = new WebSocket(external)

		___native.setUserData(external, socket)

		this.emit('connection', socket)

	}

	___onNativeSocketDisconnection (external, code, message, socket) {

		process.nextTick(() => socket.___internalOnClose(code, message))

		___native.clearUserData(external)

	}

	___onNativeSocketMessage (message, socket) {

		socket.___internalOnMessage(message)

	}

}

module.exports = {
	Client,
	Server
}


